/* generated by Makefile.devel from src/configure */
#define PACKAGE_NAME "GNU CLISP"
#define PACKAGE_TARNAME "clisp"
#define PACKAGE_VERSION "2.33 (2004-03-17)"
#define PACKAGE_STRING "GNU CLISP 2.33 (2004-03-17)"
#define PACKAGE_BUGREPORT "http://clisp.cons.org/"
